from ..files.ana import Ana
from ..files.ons import Ons
from ..files.cemaden import Cemaden
from ..files.chesf import Chesf
from ..files.nasa import Nasa
from ..files.semarh import Semarh

__all__ = ["Ana", "Cemaden", "Chesf", "Nasa", "Ons", "Semarh"]



import plotly.graph_objs as go
import numpy as np
import pandas as pd


class RatingCurve:

    def __init__(self, data: pd.DataFrame, title=None, width=None, height=None, size_text=None):
        self.data = data
        self.width = width
        self.height = height
        self.size_text = size_text
        self.title = title

    def plot(self):
        bandxaxis = go.layout.XAxis(title="% do tempo que a vazão é igualada ou superada")
        bandyaxis = go.layout.YAxis(title="Vazão(m³/s)")

        layout = dict(
            title=dict(text=self.title, x=0.5, xanchor='center', y=0.95, yanchor='top',
                       font=dict(family='Courier New, monospace', size=self.size_text + 10)),
            xaxis=bandxaxis,
            yaxis=bandyaxis,
            width=self.width, height=self.height,
            font=dict(family='Courier New, monospace', size=self.size_text, color='rgb(0,0,0)'),
            showlegend=False, plot_bgcolor='#FFFFFF', paper_bgcolor='#FFFFFF')

        data = list()
        data.append(self.data_plot(color='rgb(0,0,0)'))
        fig = dict(data=data, layout=layout)
        return fig, data

    def data_plot(self, color):
        data_y = np.sort(self.data.values)
        count = 0
        data_x = []
        for _ in data_y:
            count += 1
            data_x.append((1-(count/len(data_y)))*100)

        data = go.Scatter(x=data_x,
                          y=data_y,
                          line=dict(width=1, color=color),
                          opacity=1, connectgaps=False)
        return data



import numpy as np
from hidrocomp import statistic as e
from hidrocomp.statistic.stats_build import StatsBuild
from scipy.stats import norm


class Normal(StatsBuild):
    name = 'NOR'
    estimador = None
    parameter = {'loc': None, 'scale': None}

    def __init__(self, data: list = None, loc=None, scale=None):
        self.loc = loc
        self.scale = scale
        self.parameter['loc'] = self.loc
        self.parameter['scale'] = self.scale
        super().__init__(data, loc, scale)
        try:
            self.dist = norm(loc=self.loc, scale=self.scale)
        except TypeError:
            self.dist = None

    def z_score(self, q: float) -> float:
        if self.data is not None:
            mean = np.mean(self.data)
            std = np.std(self.data)
        else:
            mean = self.loc
            std = self.scale

        return (q - mean) / std

    def mvs(self):
        if self.data is None:
            raise e.DataNotExist("Data not's None", 35)

        mvs = norm.fit(self.data)
        self.estimador = 'MML'
        self.loc = mvs[0]
        self.scale = mvs[1]
        self.dist = norm(loc=self.loc, scale=self.scale)

        return self.loc, self.scale

    def mml(self):
        if self.data is None:
            raise e.DataNotExist("Data not's None", 39)

        mml = norm.lmom_fit(self.data)
        self.estimador = 'MML'
        self.loc = mml['loc']
        self.scale = mml['scale']
        self.dist = norm(loc=self.loc, scale=self.scale)

        return self.loc, self.scale

    def rvs(self, n):
        pass


from abc import ABCMeta, abstractmethod
import pandas as pd
import numpy as np
import calendar as cal
from hidrocomp.eflow.exceptions import *
from hidrocomp.eflow.rva import RVA
from hidrocomp.eflow.era import EraVariable, EraAspect


class Aspect(metaclass=ABCMeta):
    name = None

    def __init__(self, flow, month_start, central_metric, variation_metric, status, variables: list = None, *args,
                 **kwargs):
        self._era = None
        self.flow = flow
        self.station = flow.station
        self.month_start = month_start
        self.central_metric = central_metric
        self.variation_metric = variation_metric
        self.status = status
        self.data = self._data()
        self.metrics = self.metric_stats()
        self.args = args
        self.kwargs = kwargs

    def variable(self, name=None):
        name = name.capitalize()
        if name is None:
            raise VariableError("Variable is None")
        else:
            try:
                if self.variables[name] is None:
                    if self.status == "pre":
                        self.variables[name] = PreVariable(data=self.data[name], value=self.metrics.loc[name],
                                                           name=name)
                    elif self.status == "pos":
                        self.variables[name] = PosVariable(data=self.data[name], value=self.metrics.loc[name],
                                                           name=name)
                    else:
                        raise StatusError("Status invalid!")
                return self.variables[name]
            except KeyError:
                raise VariableError("Variable invalid! Options: {}".format(self.variables.keys()))

    def metric_stats(self):
        if self.central_metric == 'mean':
            mean = pd.DataFrame(self.data.mean(), columns=['Means'])
        elif self.central_metric == 'median':
            mean = pd.DataFrame(self.data.median(), columns=['Means'])
        else:
            raise NotStatistic('Not exist statistic {}: use {} or {}'.format(self.central_metric, 'mean', 'median'),
                               line=40)
        if self.variation_metric == 'cv':
            cv = pd.DataFrame(self.data.std() / self.data.mean(), columns=['Coeff. of Var.'])
        elif self.variation_metric == 'std':
            cv = pd.DataFrame(self.data.std(), columns=['Coeff. of Var.'])
        else:
            raise NotStatistic('Not exist statistic {}: use {} or {}'.format(self.variation_metric, 'cv', 'std'),
                               line=46)

        stats = mean.combine_first(cv)
        stats = stats.reindex(self.variables)
        return stats

    @abstractmethod
    def _data(self):
        pass

    def rva_frequency(self, aspect_pos, statistic="non-parametric", boundaries=17):
        frequency_aspect = pd.DataFrame()
        if aspect_pos.status == "pos":
            for i in self.variables:
                frequency = self.variable(name=i).rva(aspect_pos.variable(name=i), statistic=statistic,
                                                      boundaries=boundaries).frequency_pos
                frequency_aspect = frequency_aspect.combine_first(frequency)
                frequency_aspect = frequency_aspect.reindex(self.variables)
            return frequency_aspect
        else:
            raise StatusError("")

    def rva_measure_hydrologic_alteration(self, aspect_pos, statistic="non-parametric", boundaries=17):
        measure_hydrologic_alteration = pd.DataFrame()
        if aspect_pos.status == "pos":
            for i in self.variables:
                mha = self.variable(name=i).rva(aspect_pos.variable(name=i), statistic=statistic,
                                                boundaries=boundaries).measure_hydrologic_alteration()
                measure_hydrologic_alteration = measure_hydrologic_alteration.combine_first(mha)
                measure_hydrologic_alteration = measure_hydrologic_alteration.reindex(self.variables)
            return measure_hydrologic_alteration
        else:
            raise StatusError("Aspect status must be pos")

    def era(self, aspect_pos, m: int, interval: int):
        if self._era is None:
            self._era = EraAspect(name=self.name)
            if aspect_pos.status == "pos":
                for i in self.variables:
                    cha_variable = self.variable(name=i).era(variable_pos=aspect_pos.variable(name=i), m=m,
                                                             interval=interval)
                    self._era.variables = cha_variable
            else:
                raise StatusError("Aspect status must be pos")
        return self._era


class Variable:

    def __init__(self, data, value, name):
        self.data = data
        self.value = value
        self.name = name


class PosVariable(Variable):
    pass


class PreVariable(Variable):

    _era: EraVariable = None
    _rva: RVA = None

    def rva(self, variable_pos, statistic="non-parametric", boundaries=17) -> RVA:
        if self._rva is None:
            self._rva = RVA(variable_pre=self, variable_pos=variable_pos, boundaries=boundaries, statistic=statistic)
        return self._rva

    def era(self, variable_pos, m: int, interval: int) -> EraVariable:
        if self._era is None:
            self._era = EraVariable(variable_pre=self, variable_pos=variable_pos, m=m, interval=interval)
        return self._era


class Magnitude(Aspect):
    name = "Magnitude"

    variables_all = {'January': None, 'February': None, 'March': None, 'April': None, 'May': None, 'June': None,
                     'July': None, 'August': None, 'September': None, 'October': None, 'November': None,
                     'December': None}

    def __init__(self, flow, month_start, central_metric, variation_metric, status, variables: list = None):
        self.variables = {}
        if variables is not None:
            for variable in variables:
                self.variables[variable] = None
        else:
            self.variables = self.variables_all.copy()

        super().__init__(flow=flow, month_start=month_start, central_metric=central_metric,
                         variation_metric=variation_metric, status=status)

    def _data(self):

        years = self.flow.data.groupby(pd.Grouper(freq=self.month_start[1]))
        data = pd.DataFrame(columns=['January', 'February', 'March', 'April', 'May', 'June', 'July', 'August',
                                     'September', 'October', 'November', 'December'])
        for year in years:
            aux = year[1].groupby(pd.Grouper(freq='M')).mean()
            df = pd.DataFrame({year[0].year: {
                cal.month_name[i.month]: aux[self.station][i] for i in aux[self.station].index}})

            data = data.combine_first(df.T)

        mean_months = data
        return mean_months


class MagnitudeDuration(Aspect):
    name = "Magnitude and Duration"

    variables_all = {"1-day minimum": None, "1-day maximum": None, "3-day minimum": None, "3-day maximum": None,
                      "7-day minimum": None, "7-day maximum": None, "30-day minimum": None, "30-day maximum": None,
                      "90-day minimum": None, "90-day maximum": None, "Number of zero days": None,
                      "Base flow index": None}

    def __init__(self, flow, month_start, central_metric, variation_metric, status, variables: list = None,
                 events_high=None, events_low=None):
        self.events_high = events_high
        self.events_low = events_low

        self.variables = {}
        if variables is not None:
            for variable in variables:
                self.variables[variable] = None
        else:
            self.variables = self.variables_all.copy()
        super().__init__(flow=flow, month_start=month_start, central_metric=central_metric,
                         variation_metric=variation_metric, status=status)

    def _data(self):
        if self.events_high is None is self.events_low:
            aver_data = pd.DataFrame()
            for i in [1, 3, 7, 30, 90]:
                ave_max = self.flow.data.rolling(window=i).mean().groupby(pd.Grouper(freq=self.month_start[1])).max()
                ave_min = self.flow.data.rolling(window=i).mean().groupby(pd.Grouper(freq=self.month_start[1])).min()
                years = ave_max.index.year
                df1 = pd.DataFrame(pd.Series(data=ave_min[self.station].values, name='%s-day minimum' % i,
                                             index=ave_max.index.year))
                df2 = pd.DataFrame(pd.Series(data=ave_max[self.station].values, name='%s-day maximum' % i,
                                             index=ave_min.index.year))
                aver_data = aver_data.combine_first(df1)
                aver_data = aver_data.combine_first(df2)
                if i == 7:
                    mean_year = self.flow.data[self.station].groupby(pd.Grouper(freq=self.month_start[1])).mean().values
                    base_flow = pd.DataFrame(pd.Series(data=ave_min[self.station].values / mean_year,
                                                       name='Base flow index', index=years))
                    aver_data = aver_data.combine_first(base_flow)

            dic_zero = {i[0].year: i[1].loc[i[1][self.station].values == 0].sum().values[0]
                        for i in self.flow.data.groupby(pd.Grouper(freq=self.month_start[1]))}
            serie_dict_zero = pd.Series(data=dic_zero, name='Number of zero days')
            serie_dict_zero.loc[serie_dict_zero.isnull()].value = 0
            if serie_dict_zero.sum() > 0:
                magn_and_duration = aver_data.combine_first(pd.DataFrame(serie_dict_zero))
            else:
                if 'Number of zero days' in self.variables.keys():
                    del self.variables['Number of zero days']
                magn_and_duration = aver_data
        else:
            peaks_high = self.events_high.peaks
            if len(peaks_high) > 0:
                peaks_high.index.name = None
                maximum = peaks_high["Peaks"].groupby(pd.Grouper(freq=self.month_start[1])).mean()
                maximum.index = maximum.index.year
                maximum.name = "1-day maximum"
            else:
                maximum = pd.Series(name="1-day maximum")

            peaks_low = self.events_low.peaks
            if len(peaks_low) > 0:
                peaks_low.index.name = None
                minimum = peaks_low["Peaks"].groupby(pd.Grouper(freq=self.month_start[1])).mean()
                minimum.index = minimum.index.year
                minimum.name = "1-day minimum"
            else:
                minimum = pd.Series(name="1-day minimum")
            magn_and_duration = minimum.to_frame().combine_first(maximum.to_frame())
        return magn_and_duration


class TimingExtreme(Aspect):
    name = "Timing Extreme"
    variables_all = {"Date of minimum": None, "Date of maximum": None}

    def __init__(self, flow, month_start, central_metric, variation_metric, status, variables: list = None,
                 events_high=None, events_low=None):
        self.events_high = events_high
        self.events_low = events_low

        self.variables = {}
        if variables is not None:
            for variable in variables:
                self.variables[variable] = None
        else:
            self.variables = self.variables_all.copy()
        super().__init__(flow=flow, month_start=month_start, central_metric=central_metric,
                         variation_metric=variation_metric, status=status)

    def _data(self):
        if self.events_high is None is self.events_low:
            day_julian_max = self.flow.data[self.station].groupby(
                pd.Grouper(freq=self.month_start[1])).idxmax().dropna(axis=0)
            day_julian_min = self.flow.data[self.station].groupby(
                pd.Grouper(freq=self.month_start[1])).idxmin().dropna(axis=0)

            df_day_julian_max = pd.DataFrame(list(map(int, pd.DatetimeIndex(day_julian_max.values).strftime("%j"))),
                                             index=day_julian_max.index.year,
                                             columns=["Date of maximum"])

            df_day_julian_min = pd.DataFrame(list(map(int, pd.DatetimeIndex(day_julian_min.values).strftime("%j"))),
                                             index=day_julian_min.index.year,
                                             columns=["Date of minimum"])
        else:
            peaks_high = self.events_high.peaks
            if len(peaks_high) > 0:
                peaks_high.index.name = None
                df_day_julian_max = pd.DataFrame(data=peaks_high.index.strftime("%j"), index=peaks_high.index,
                                                 columns=["Date of maximum"],
                                                 dtype="float64").groupby(pd.Grouper(freq=self.month_start[1])).mean()
                df_day_julian_max.index = df_day_julian_max.index.year
            else:
                df_day_julian_max = pd.DataFrame(columns=["Date of maximum"])

            peaks_low = self.events_low.peaks
            if len(peaks_low) > 0:
                peaks_low.index.name = None
                df_day_julian_min = pd.DataFrame(data=peaks_low.index.strftime("%j"), index=peaks_low.index,
                                                 columns=["Date of minimum"],
                                                 dtype="float64").groupby(pd.Grouper(freq=self.month_start[1])).mean()
                df_day_julian_min.index = df_day_julian_min.index.year
            else:
                df_day_julian_min = pd.DataFrame(columns=["Date of minimum"])

        # combine the dfs of days julian
        timing_extreme = df_day_julian_max.combine_first(df_day_julian_min)
        # print(timing_extreme)
        return timing_extreme


class FrequencyDuration(Aspect):
    name = "Frequency and Duration"
    variables_all = {"High events count": None, "Low events count": None, "High events duration": None,
                     "Low events duration": None}

    def __init__(self, flow, month_start, central_metric, variation_metric, status, events_high, events_low,
                 variables: list = None, *args, **kwargs):
        self.events_high = events_high
        self.events_low = events_low
        self.args = args
        self.kwargs = kwargs
        self.variables = {}
        if variables is not None:
            for variable in variables:
                self.variables[variable] = None
        else:
            self.variables = self.variables_all.copy()
        super().__init__(flow=flow, month_start=month_start, central_metric=central_metric,
                         variation_metric=variation_metric, status=status)

    def _data(self):

        def aux_frequency_and_duration(events):
            name = {'flood': 'High', 'drought': 'Low'}
            type_event = events.type_event

            group = pd.DataFrame(
                index=pd.date_range(events.obj.date_start, events.obj.date_end, freq=self.month_start[1]),
                columns=["{} events duration".format(name[type_event]),
                         '{} events count'.format(name[type_event])])

            if len(events.peaks) != 0:
                duration_pulse = pd.DataFrame(events.peaks.groupby(
                    pd.Grouper(freq=self.month_start[1])).Duration.mean()).rename(
                    columns={"Duration": '{} events duration'.format(name[type_event])})

                pulse = pd.DataFrame(events.peaks.resample(self.month_start[1]).count().Duration).rename(
                    columns={"Duration": '{} events count'.format(name[type_event])})

                group = group.combine_first(duration_pulse)
                group = group.combine_first(pulse)

            group = group.fillna(value={'{} events count'.format(name[type_event]): 0})
            threshold = pd.DataFrame(pd.Series(events.threshold, name="{} Threshold".format(name[type_event])))
            return group, threshold

        frequency_and_duration_high, threshold_high_mag = aux_frequency_and_duration(self.events_high)

        frequency_and_duration_low, threshold_low_mag = aux_frequency_and_duration(self.events_low)

        frequency_and_duration = frequency_and_duration_high.combine_first(frequency_and_duration_low)
        frequency_and_duration.index = frequency_and_duration.index.year
        return frequency_and_duration


class RateFrequency(Aspect):
    name = "Rate and Frequency"
    variables_all = {"Rise rate": None, "Fall rate": None, "Number of reversals": None}

    def __init__(self, flow, month_start, central_metric, variation_metric, status, variables: list = None):
        self.variables = {}
        if variables is not None:
            for variable in variables:
                self.variables[variable] = None
        else:
            self.variables = self.variables_all.copy()
        super().__init__(flow=flow, month_start=month_start, central_metric=central_metric,
                         variation_metric=variation_metric, status=status)

    # <editor-fold desc="Check type rate"
    @staticmethod
    def check_rate(value1, value2, type_rate):
        if type_rate == 'Rise rate':
            return value1 < value2
        elif type_rate == 'Fall rate':
            return value1 > value2
    # </editor-fold>

    def _data(self):
        # self.variables = {"Rise rate": None, "Fall rate": None, "Number of reversals": None}

        data_water = self.flow.data.groupby(pd.Grouper(freq=self.month_start[1]))
        rate_df = pd.DataFrame(columns=['Rise rate', 'Fall rate', 'Number of reversals'])
        boo = False
        mean = 0
        for type_rate in ['Rise rate', 'Fall rate']:
            for key, series in data_water:
                d1 = None
                values = []
                cont = 0
                for i in series.index:
                    if d1 is not None:
                        if self.check_rate(self.flow.data.loc[d1, self.station], self.flow.data.loc[i, self.station],
                                           type_rate):
                            boo = True
                            values.append(self.flow.data.loc[i, self.station] - self.flow.data.loc[d1, self.station])
                        else:
                            if boo:
                                mean = np.mean(values)
                                cont += 1
                                boo = False
                    d1 = i
                if boo:
                    mean = np.mean(values)
                    boo = False

                rate_df.at[key.year, type_rate] = mean
                if rate_df['Number of reversals'][key.year] > 0:
                    rate_df.at[key.year, 'Number of reversals'] = cont + rate_df['Number of reversals'][key.year]
                else:
                    rate_df.at[key.year, 'Number of reversals'] = cont

        return rate_df



from typing import Union

from hidrocomp.eflow.exceptions import *
import pandas as pd
import calendar as cal
from hidrocomp.eflow.aspect import Magnitude, MagnitudeDuration, TimingExtreme, FrequencyDuration, RateFrequency
from hidrocomp.eflow import Era, RVA


class IHA:

    aspects_all = {"Magnitude": Union[Magnitude], "Magnitude and Duration": Union[MagnitudeDuration],
                   "Timing Extreme": Union[TimingExtreme], "Frequency and Duration": Union[FrequencyDuration],
                   "Rate and Frequency": Union[RateFrequency]}

    def __init__(self, flow, date_start: str = None, date_end: str = None, statistic="no-parametric",
                 central_metric="mean", variation_metric: str = "std", type_threshold="stationary", status=None,
                 month_water: int = None, type_criterion: str = None, threshold_high: float = None,
                 threshold_low: float = None, aspects: list = None, magnitude: list = None,
                 magnitude_and_duration: list = None, timing: list = None, frequency_and_duration: list = None,
                 rate_and_frequency: list = None, new_iha: bool = False, **kwargs):
        """
        :param data: pandas Series
        :param month_water: initial month water (int: referent the month, ex.: 1 for Jan, 2 for Fev)
        :param station: station of data
        :param status: 'pre' or 'pos'
        :param date_start: Date of start application ('dd/mm/aaaa')
        :param date_end: Date of end application ('dd/mm/aaaa')
        :param statistic: 'non-parametric or parametric'
        :param central_metric: 'mean or median'
        :param variation_metric: 'str' or 'cv'
        """

        if aspects is not None:
            self.aspects = {}
            for aspect in aspects:
                self.aspects[aspect] = self.aspects_all[aspect]
        else:
            self.aspects = self.aspects_all.copy()

        self.flow = flow
        self.status = status
        self._era = None
        self.month_start = self.get_month_start(month_water)
        self.date_start = pd.to_datetime(date_start, dayfirst=True)
        self.date_end = pd.to_datetime(date_end, dayfirst=True)
        self.statistic = statistic
        self.__events_high = None
        self.__events_low = None
        self.central_metric = central_metric
        self.variation_metric = variation_metric
        self.type_threshold = type_threshold
        self.type_criterion = type_criterion
        self.threshold_high = threshold_high
        self.threshold_low = threshold_low
        self.magnitude_variables = magnitude
        self.magnitude_and_duration_variables = magnitude_and_duration
        self.timing_extreme_variables = timing
        self.frequency_and_duration_variables = frequency_and_duration
        self.rate_and_frequency_variables = rate_and_frequency
        self.new_iha = new_iha
        self.kwargs = kwargs

    @property
    def events_high(self):
        if self.__events_high is None:
            self.__events_high = self.flow.partial(type_threshold=self.type_threshold, type_event="flood",
                                                   type_criterion=self.type_criterion,
                                                   value_threshold=self.threshold_high, **self.kwargs)
        return self.__events_high

    @property
    def events_low(self):
        if self.__events_low is None:
            self.__events_low = self.flow.partial(type_event='drought', type_threshold=self.type_threshold,
                                                  type_criterion=self.type_criterion,
                                                  value_threshold=self.threshold_low, **self.kwargs)
        return self.__events_low

    # <editor-fold desc="Return Station">
    def get_station(self, station):
        if len(self.flow.data.columns.values) != 1:
            if station is None:
                raise NotStation("Station requirement")
            else:
                if station in self.flow.data.columns.values:
                    return station
                else:
                    raise NotStation("Not station")
        else:
            get_station = self.flow.data.columns.values[0]
            self.flow.data = pd.DataFrame(self.flow.data[get_station])

            return get_station

    # </editor-fold>

    # <editor-fold desc= "Return Month Water">
    def get_month_start(self, month_water=None):
        """
        :param month_water:
        :return self.month_water:
        """
        if month_water is None:
            return self.flow.month_start_year_hydrologic()
        else:
            self.flow.month_abr = 'AS-%s' % cal.month_abbr[month_water].upper()
            self.flow.month_num = month_water
            return self.flow.month_num, self.flow.month_abr

    # </editor-fold>

    def __str__(self) -> str:
        return self.summary().__str__()

    def rva(self) -> RVA:
        pass

    def era(self, iha_obs, m: int = 500, interval: int = 95) -> Era:
        if self.status == "pos":
            raise StatusError("Dhram not available for self object!")
        if iha_obs.status == "pre":
            raise StatusError(f"status of {iha_obs} invalid!")

        self._era = Era()
        for aspect in self.aspects:
            if aspect == "Magnitude":
                self._era.aspects = self.magnitude.era(aspect_pos=iha_obs.magnitude, m=m, interval=interval)
            elif aspect == "Magnitude and Duration":
                self._era.aspects = self.magnitude_and_duration.era(aspect_pos=iha_obs.magnitude_and_duration, m=m,
                                                                    interval=interval)
            elif aspect == "Timing Extreme":
                self._era.aspects = self.timing_extreme.era(aspect_pos=iha_obs.timing_extreme, m=m, interval=interval)
            elif aspect == "Frequency and Duration":
                self._era.aspects = self.frequency_and_duration.era(aspect_pos=iha_obs.frequency_and_duration, m=m,
                                                                    interval=interval)
            elif aspect == "Rate and Frequency":
                self._era.aspects = self.rate_and_frequency.era(aspect_pos=iha_obs.rate_and_frequency, m=m,
                                                                interval=interval)
        return self._era

    def summary(self):
        df = pd.DataFrame()
        idx = []
        for aspect in self.aspects:
            if aspect == "Magnitude":
                df = df.combine_first(self.magnitude.metrics)
                idx = idx + list(self.magnitude.variables.keys())
            elif aspect == "Magnitude and Duration":
                df = df.combine_first(self.magnitude_and_duration.metrics)
                idx = idx + list(self.magnitude_and_duration.variables.keys())
            elif aspect == "Timing Extreme":
                df = df.combine_first(self.timing_extreme.metrics)
                idx = idx + list(self.timing_extreme.variables.keys())
            elif aspect == "Frequency and Duration":
                df = df.combine_first(self.frequency_and_duration.metrics)
                idx = idx + list(self.frequency_and_duration.variables.keys())
            elif aspect == "Rate and Frequency":
                df = df.combine_first(self.rate_and_frequency.metrics)
                idx = idx + list(self.rate_and_frequency.variables.keys())
        return df.reindex(idx)

    # <editor-fold desc="Group 1: Magnitude of monthly water conditions">
    @property
    def magnitude(self) -> Magnitude:
        if not isinstance(self.aspects["Magnitude"], Magnitude):
            magnit = Magnitude(flow=self.flow, month_start=self.month_start, central_metric=self.central_metric,
                               variation_metric=self.variation_metric, status=self.status,
                               variables=self.magnitude_variables)

            self.aspects["Magnitude"] = magnit
        return self.aspects["Magnitude"]

    # </editor-fold>

    # <editor-fold desc="Group 2: Magnitude and Duration of annual extreme water conditions">
    @property
    def magnitude_and_duration(self) -> MagnitudeDuration:
        if not isinstance(self.aspects["Magnitude and Duration"], MagnitudeDuration):
            if self.new_iha:
                magnit_and_durat = MagnitudeDuration(flow=self.flow, month_start=self.month_start, status=self.status,
                                                     central_metric=self.central_metric, events_high=self.events_high,
                                                     events_low=self.events_low,
                                                     variables=self.magnitude_and_duration_variables,
                                                     variation_metric=self.variation_metric)
            else:
                magnit_and_durat = MagnitudeDuration(flow=self.flow, month_start=self.month_start, status=self.status,
                                                     central_metric=self.central_metric,
                                                     variables=self.magnitude_and_duration_variables,
                                                     variation_metric=self.variation_metric)
            self.aspects["Magnitude and Duration"] = magnit_and_durat
        return self.aspects["Magnitude and Duration"]
    # </editor-fold>

    # <editor-fold desc="Group 3: Timing of annual extreme water conditions">
    @property
    def timing_extreme(self) -> TimingExtreme:
        if not isinstance(self.aspects["Timing Extreme"], TimingExtreme):
            timing = TimingExtreme(flow=self.flow, month_start=self.month_start, central_metric=self.central_metric,
                                   variation_metric=self.variation_metric, status=self.status,
                                   events_high=self.events_high, events_low=self.events_low,
                                   variables=self.timing_extreme_variables)

            self.aspects["Timing Extreme"] = timing
        return self.aspects["Timing Extreme"]
    # </editor-fold>

    # <editor-fold desc="Group 4: Frequency and duration of high and low pulses">
    @property
    def frequency_and_duration(self) -> FrequencyDuration:
        if not isinstance(self.aspects["Frequency and Duration"], FrequencyDuration):
            freq = FrequencyDuration(flow=self.flow, month_start=self.month_start, central_metric=self.central_metric,
                                     variation_metric=self.variation_metric, status=self.status,
                                     events_high=self.events_high, events_low=self.events_low,
                                     variables=self.frequency_and_duration_variables, **self.kwargs)

            self.aspects["Frequency and Duration"] = freq
        return self.aspects["Frequency and Duration"]
    # </editor-fold>

    # <editor-fold desc="Group 5: Rate and frequency of water condition changes">
    @property
    def rate_and_frequency(self) -> RateFrequency:
        if not isinstance(self.aspects["Rate and Frequency"], RateFrequency):
            rate = RateFrequency(flow=self.flow, month_start=self.month_start, central_metric=self.central_metric,
                                 variation_metric=self.variation_metric, status=self.status,
                                 variables=self.rate_and_frequency_variables)

            self.aspects["Rate and Frequency"] = rate
        return self.aspects["Rate and Frequency"]
    # </editor-fold>

    @property
    def aspects_name(self) -> str:
        return f"Aspects_names({list(self.aspects.keys())})"



import pandas as pd

from hidrocomp.eflow.exceptions import *
from hidrocomp.eflow.graphics import Graphics, GraphicsRVA


class RVA:

    def __init__(self, variable_pre, variable_pos, statistic, boundaries):
        self.data_pre = variable_pre.data
        self.data_pos = variable_pos.data
        self.boundaries = boundaries
        self.statistic = statistic
        self.name_variable = variable_pre.name
        self.line = self.__line()
        self.frequency_pre = self.__frequency(data_variable=self.data_pre)
        self.frequency_pos = self.__frequency(data_variable=self.data_pos)

    def measure_hydrologic_alteration(self):
        def very(rva):
            for i in rva:
                for j in rva[i].index:
                    if rva[i][j] < -1:
                        rva.iat[i, j] = -1
            return rva
        try:
            mha = (self.frequency_pos - self.frequency_pre) / self.frequency_pre
        except ZeroDivisionError:
            mha = (self.frequency_pos - self.frequency_pre)
        return very(mha)

    def __line(self):
        line = pd.DataFrame(columns=['lower_line', 'upper_line', 'median_line'])

        if self.statistic == 'non-parametric':
            line.at[self.name_variable, 'lower_line'] = self.data_pre.quantile((50 - self.boundaries) / 100)
            line.at[self.name_variable, 'upper_line'] = self.data_pre.quantile((50 + self.boundaries) / 100)
            line.at[self.name_variable, 'median_line'] = self.data_pre.median()
            return line
        elif self.statistic == 'parametric':
            line.at[self.name_variable,
                    'lower_line'] = self.data_pre[self.name_variable].mean() - self.data_pre[self.name_variable].std()
            line.at[self.name_variable,
                    'lower_line'] = self.data_pre[self.name_variable].mean() + self.data_pre[self.name_variable].std()
            line.at[self.name_variable, 'lower_line'] = self.data_pre[self.name_variable].median()
            return line
        else:
            raise NotStatistic('Not exist statistic {}: use {} or {}'.format(self.statistic,
                                                                             'non-parametric', 'parametric'), line=91)

    def __frequency(self, data_variable):
        count = pd.DataFrame(columns=['Lower', 'Median', 'Upper'])
        upper_line, lower_line = self.line['upper_line'], self.line['lower_line']
        if upper_line[self.name_variable] == 0 and lower_line[self.name_variable] == 0:
            count.at[self.name_variable, 'Lower'] = 0
            count.at[self.name_variable, 'Upper'] = 0
            count.at[self.name_variable, 'Median'] = 0
        else:
            boolean_lower = pd.DataFrame(data_variable.isin(
                data_variable.loc[data_variable <= lower_line[self.name_variable]]))
            boolean_upper = pd.DataFrame(data_variable.isin(
                data_variable.loc[data_variable >= upper_line[self.name_variable]]))
            boolean_median = pd.DataFrame(data_variable.isin(
                data_variable.loc[boolean_lower[self.name_variable] == boolean_upper[self.name_variable]]))

            count.at[self.name_variable, 'Lower'] = boolean_lower[
                self.name_variable].loc[boolean_lower[self.name_variable] == True].count()
            count.at[self.name_variable, 'Upper'] = boolean_upper[
                self.name_variable].loc[boolean_upper[self.name_variable] == True].count()
            count.at[self.name_variable, 'Median'] = boolean_median[
                self.name_variable].loc[boolean_median[self.name_variable] == True].count()
        return count

    def plot(self, color={"pre": "blue", "pos": "red"}):
        """
        @type color: dict
        """
        fig_obs, data_obs = GraphicsRVA(data_variable=self.data_pos, status="pos", color=color, line=self.line).plot()
        fig_nat, data_nat = GraphicsRVA(data_variable=self.data_pre, status="pre", color=color, line=self.line).plot()

        data = data_obs + [data_nat[0]]
        fig = dict(data=data, layout=fig_nat['layout'])

        return fig, data



from .rating_curve import RatingCurve
from .boxplot import Boxplot
from .genextreme import GenExtreme
from .hydrogram_annual import HydrogramAnnual
from .hydrogram_by_year import HydrogramYear
from .hydrogram_parcial import HydrogramParcial
from .hydrogram_clean import HydrogramClean
from .gantt import Gantt
from .genpareto import GenPareto
from .polar import Polar

__all__ = ['GenPareto', 'GenExtreme', 'HydrogramYear', 'HydrogramParcial', 'HydrogramAnnual', 'Gantt', 'Polar',
           'RatingCurve', 'Boxplot', 'HydrogramClean']



import pandas as pd
import math
import scipy.stats as stat
import plotly as py
import plotly.figure_factory as FF

from hidrocomp.statistic.genpareto import Gpa
from hidrocomp.graphics.gantt import Gantt
from hidrocomp.graphics.genpareto import GenPareto
from hidrocomp.graphics.hydrogram_parcial import HydrogramParcial
from hidrocomp.graphics.boxplot import Boxplot
from hidrocomp.graphics.polar import Polar


class Partial(object):
    distribution = 'GPA'
    __percentil = 0.8
    dic_name = {'stationary': 'Percentil', 'events_by_year': 'Eventos por Ano', 'autocorrelation': 'Autocorrelação'}

    def __init__(self, obj, station, type_threshold, value_threshold, type_event, type_criterion, **kwargs):
        """
            Parameter:
                obj: Object Series;
                station: Point observation of date
                type_threshold: Tipo de calculo do limiar('stationary' or
                                                          'events_by_year')
                value_threshold: Valor do limiar:
                    Para type_threshold = 'stationary': percentil ou valor
                    Para type_threshold = 'events_by_year': quantidade média de
                                                            picos por ano
                type_criterion: Critério de Independência ('mean', 'median',
                                'xmin_bigger_dois_terco_x', 'xmin_bigger_qmin')
                type_event: Tipo do evento em estudo ('flood' or 'drought')
        """
        self.obj = obj
        self.data = self.obj.data
        self.station = station
        self.__peaks = None
        self.threshold_criterion = None
        self.fit = None
        self.type_threshold = type_threshold
        self.type_criterion = type_criterion
        self.type_event = type_event
        self.value = value_threshold

        if self.type_criterion == 'median':
            self.__percentil = 0.65
        elif self.type_criterion == 'autocorrelation':
            self.duration = kwargs['duration']
        elif self.type_criterion == "wrc":
            self.duration = kwargs['duration']


        self.__threshold(self.value)
        if self.type_criterion is not None:
            self.name = '%s(%s) - %s' % (self.dic_name[self.type_threshold], self.value, self.type_criterion.title())
        else:
            self.name = '%s(%s)' % (self.dic_name[self.type_threshold], self.value)

        if self.peaks is None:
            self.dist_gpa = Gpa(data=self.peaks["Peaks"])

    @property
    def peaks(self):
        if self.__peaks is None:
            if self.type_criterion == "wrc":
                if self.type_event == "flood":
                    self.__peaks = self.__criterion_water_resources_council()
                elif self.type_event == "drought":
                    self.__peaks = self.__events_over_threshold()
                    if len(self.__peaks) > 0:
                        self.__peaks.at[self.__peaks.index[-1],
                                        "End"] = self.__peaks["End"].iloc[-1] - pd.to_timedelta(1, unit="d")

            elif self.type_criterion == "duration":
                if self.type_event == "flood":
                    self.__peaks = self.__duration()
                elif self.type_event == "drought":
                    self.__peaks = self.__events_over_threshold()
                    self.__peaks.at[self.__peaks.iloc[-1].index,
                                    "End"] = self.__peaks.iloc[-1].End - pd.timedelta_range(start='1 day', periods=1,
                                                                                            freq='D')
            elif self.type_criterion == "autocorrelation":
                if self.type_event == "flood":
                    self.__peaks = self.__test_autocorrelation()
                elif self.type_event == "drought":
                    self.__peaks = self.__test_autocorrelation()
                    if len(self.__peaks) > 0:
                        self.__peaks.at[self.__peaks.sort_values(by="End").index[-1],
                                        "End"] = self.__peaks.sort_values(
                            by="End")["End"].iloc[-1] - pd.to_timedelta(1, unit="d")
            else:
                if self.type_event == "flood":
                    self.__peaks = self.__events_over_threshold()
                elif self.type_event == "drought":
                    self.__peaks = self.__events_over_threshold()
                    self.__peaks.at[self.__peaks.sort_values(by="End").index[-1],
                                    "End"] = self.__peaks.sort_values(
                        by="End")["End"].iloc[-1] - pd.to_timedelta(1, unit="d")
        return self.__peaks

    def __events_over_threshold(self):

        if self.type_event == 'flood':
            data = self.data.loc[self.data[self.station] > self.threshold, self.station].to_frame()

        elif self.type_event == 'drought':
            data = self.data.loc[self.data[self.station] < self.threshold, self.station].to_frame()

        else:
            raise TypeError("Type events {} invalid! Use flood or drought".format(self.type_event))

        date_start, date_end = self.__start_and_end(data=data)
        if date_start is not None and date_end is not None:
            _data = pd.DataFrame(index=pd.date_range(start=date_start, end=date_end))
            data = _data.combine_first(data[date_start:date_end])

        max_events = {'Date': list(), 'Peaks': list(), 'Start': list(), 'End': list(), "Duration": list()}
        events = self.__period_events(data=data, station=self.station)
        for i in events.index:
            start = events["Start"][i]
            end = events["Finish"][i]
            duration = (end - start)
            if self.type_event == "flood":
                peaks = data[start:end].max()[0]
                date_max = data[start:end].idxmax()[0]
            else:
                peaks = data[start:end].min()[0]
                date_max = data[start:end].idxmin()[0]

            max_events["Date"].append(date_max)
            max_events["Peaks"].append(peaks)
            max_events["Start"].append(start)
            end = end + pd.timedelta_range(start='1 day', periods=1, freq='D')
            max_events["End"].append(end.values[0])
            max_events["Duration"].append(duration.days + 1)

        df = pd.DataFrame(data=max_events, index=max_events["Date"], columns=['Peaks', 'Start', 'End', 'Duration'])
        return df

    def __duration(self):
        events_over_threshold = self.__events_over_threshold().sort_values(by="Peaks", ascending=False)

        events_duration = pd.DataFrame()

        for i in events_over_threshold.index:
            actual = events_over_threshold.loc[events_over_threshold.index == i]

            if len(events_duration) == 0:
                events_duration = events_duration.combine_first(actual).sort_values(by="Peaks", ascending=False)
                events_duration.at[i, "Date"] = i
            else:
                if_add = False
                for j in events_duration.index:
                    duration = i - j
                    if abs(duration.days) <= self.duration:
                        events_duration = self.__update_peaks(actual=actual, events=events_duration,
                                                              idx=j).sort_values(by="Peaks", ascending=False)
                        if_add = True
                        break
                if not if_add:
                    events_duration = events_duration.combine_first(actual).sort_values(by="Peaks", ascending=False)
                    events_duration.at[i, "Date"] = i

        if len(events_duration) > 0:
            events_duration = events_duration.set_index("Date")

        return events_duration

    def __threshold(self, value):
        if value > 1 and self.type_threshold == 'events_by_year':
            self.threshold = self.data[self.station].quantile(self.__percentil)
        elif value > 1 and self.type_threshold == 'stationary':
            self.threshold = value
        else:
            self.threshold = self.data[self.station].quantile(value)
        return self.threshold

    def __test_autocorrelation(self):
        peaks = self.__duration()
        try:
            n = len(peaks.Peaks)
            serie = pd.Series(peaks.Peaks, index=peaks.index)
            self.lag1 = serie.autocorr(lag=1)
            self.lag2 = serie.autocorr(lag=2)
            r11_n = (-1 + 1.645 * math.sqrt(n - 1 - 1)) / (n - 1)
            r12_n = (-1 - 1.645 * math.sqrt(n - 1 - 1)) / (n - 1)
            r21_n = (-1 + 1.645 * math.sqrt(n - 2 - 1)) / (n - 2)
            r22_n = (-1 - 1.645 * math.sqrt(n - 2 - 1)) / (n - 2)

            if r11_n > self.lag1 > r12_n and r21_n > self.lag2 > r22_n:
                self.duration += 1
                return self.__test_autocorrelation()
            return peaks
        except (AttributeError, ValueError):
            return peaks

    @staticmethod
    def __update_peaks(actual, events, idx):
        if actual["End"].values[0] > events["End"][idx]:
            end = actual["End"].values[0]
        else:
            end = events["End"][idx]

        if actual["Start"].values[0] > events["Start"][idx]:
            start = events["Start"][idx]
        else:
            start = actual["Start"].values[0]

        events["Start"][idx] = start
        events["End"][idx] = end

        duration_into_events = end - start
        events["Duration"][idx] = duration_into_events.days

        peaks = [actual["Peaks"].values[0], events["Peaks"][idx]]
        date_peaks = [actual.index.values[0], events["Date"][idx]]
        peak_idx_max = peaks.index(max(peaks))
        date = date_peaks[peak_idx_max]
        peaks = peaks[peak_idx_max]
        events["Peaks"][idx] = peaks
        events.at[idx, "Date"] = date

        return events

    def __criterion_water_resources_council(self):
        """
        Events dependents
        theta < 5days + log(A) or (3/4)*min[Q1, Q2]
        """
        try:
            events_duration = self.__duration().sort_values(by="Peaks", ascending=False)

            events_wrc = pd.DataFrame()
            for i in events_duration.index:
                actual = events_duration.loc[events_duration.index == i]

                if len(events_wrc) == 0:
                    events_wrc = events_wrc.combine_first(actual).sort_values(by="Peaks", ascending=False)
                    events_wrc.at[i, "Date"] = i

                else:
                    if_add = False
                    for j in events_wrc.index:
                        date_range = pd.date_range(j, i)
                        qmin = self.data[self.station].loc[date_range].min()
                        q1 = events_wrc["Peaks"][j]
                        q2 = actual["Peaks"][i]

                        if qmin > (3/4)*min([q1, q2]):
                            events_wrc = self.__update_peaks(actual=actual, events=events_wrc,
                                                             idx=j).sort_values(by="Peaks", ascending=False)
                            if_add = True
                            break
                    if not if_add:
                        events_wrc = events_wrc.combine_first(actual).sort_values(by="Peaks", ascending=False)
                        events_wrc.at[i, "Date"] = i

            if len(events_wrc) > 0:
                events_wrc = events_wrc.set_index("Date")
        except KeyError:
            events_wrc = self.__duration()
        return events_wrc

    @staticmethod
    def __start_and_end(data):
        if len(data) == 0:
            return None, None
        try:
            boolean = data.dropna(axis=0, how='all')
        except AttributeError:
            boolean = data
        date = boolean.index
        return date[0], date[-1]

    @staticmethod
    def __period_events(data, station):
        """
        """
        aux = list()
        list_start = list()
        list_end = list()
        gantt_bool = data[station].isnull()
        for i in gantt_bool.index:
            if ~gantt_bool.loc[i]:
                aux.append(i)
            elif len(aux) >= 1 and gantt_bool.loc[i]:
                list_start.append(aux[0])
                list_end.append(aux[-1])
                aux = []
        if len(aux) > 0:
            list_start.append(aux[0])
            list_end.append(aux[-1])

        dic = {'Start': list_start, 'Finish': list_end}
        return pd.DataFrame(dic)

    def magnitude(self, period_return, estimador):
        if estimador == 'MML':
            self.dist_gpa.mml()
        elif estimador == 'MVS':
            self.dist_gpa.mvs()

        p = 1 - (1 / period_return)
        return self.dist_gpa.values(p)

    def period_return(self, magnitude, estimador):
        if estimador == 'MML':
            self.dist_gpa.mml()
        elif estimador == 'MVS':
            self.dist_gpa.mvs()

        p = self.dist_gpa.probs(magnitude)
        return 1 / (1 - p)

    def plot_distribution(self, title, type_function, save=False):
        parameter = self.dist_gpa.parameter
        genpareto = GenPareto(title, shape=parameter["shape"], location=parameter["loc"], scale=parameter["scale"])
        data, fig = genpareto.plot(type_function)
        if save:
            aux_name = title.replace(' ', '_')
            py.image.save_as(fig, filename='gráficos/' + '%s.png' % aux_name)
        return data, fig

    def plot_spells(self, title, size_text=14):
        if self.type_event == "drought":
            month_start_abr, month_start_num = self.obj.month_abr_drought, self.obj.month_num_drought
        elif self.type_event == "flood":
            month_start_abr, month_start_num = self.obj.month_abr_flood, self.obj.month_num_flood
        else:
            raise TypeError

        df_spells, df, month_start, month_end = Gantt.get_spells(data_peaks=self.peaks,
                                                                 month_water=[month_start_num, month_start_abr])

        df_spells = df_spells.sort_values('Task', ascending=False).reset_index(drop=True)

        colors = '#000000'
        fig = FF.create_gantt(df_spells, group_tasks=True, colors=colors, title=title, height=900, width=1200)

        fig['layout'].update(autosize=True)
        fig['layout']['xaxis'].update(title="Month", range=[month_start, month_end], tickformat="%b")
        fig['layout']['yaxis'].update(title="Year")
        fig['layout']['xaxis']['rangeselector'] = {}
        fig.layout.title = dict(text=title, x=0.5, xanchor='center', y=0.9, yanchor='top',
                                font=dict(family='Courier New, monospace', color='#7f7f7f', size=size_text + 6))
        fig.layout.font = dict(family='Courier New, monospace', size=size_text, color='#7f7f7f')
        fig.layout.plot_bgcolor = 'rgba(0,0,0,0)'
        return fig, df_spells

    def plot_polar(self, title=None, width=900, height=900, size_text=14, color=None, name=None):
        if self.type_event == 'flood':
            if title is None:
                title = 'Maximum Partial'
            elif name is None:
                name = 'Maximum peaks'
        elif self.type_event == 'drought':
            if title is None:
                title = 'Minimum Partial'
            elif name is None:
                name = 'Minimum peaks'

        _polar = Polar(df_events=self.peaks)
        fig, data = _polar.plot(width=width, height=height, size_text=size_text, title=title, color=color, name=name)

        return fig, data

    def plot_hydrogram(self, title, width=None, height=None, size_text=16, color=None, line_threshold: bool = True,
                       point_start_end: bool = True):
        hydrogram = HydrogramParcial(data=self.data, peaks=self.peaks, threshold=self.threshold, station=self.station,
                                     threshold_criterion=self.threshold_criterion, title=title, width=width,
                                     type_criterion=self.type_criterion, height=height, size_text=size_text,
                                     color=color, line_threshold=line_threshold, point_start_end=point_start_end)
        fig, data = hydrogram.plot()
        return fig, data



import pandas as pd
import plotly.graph_objs as go
from hidrocomp.statistic.normal import Normal
from hidrocomp.statistic.bootstrap import Bootstrap
from hidrocomp.eflow.exceptions import *
from hidrocomp.eflow.graphics import GraphicsCha


class Era:

    def __init__(self):
        self._point = None
        self._aspects: dict = {}
        self._classification = None
        self._abnormality = None

    @property
    def aspects(self):
        return self._aspects

    @aspects.setter
    def aspects(self, aspect):
        self._aspects[aspect.name] = aspect

    def values_mean(self):
        df = pd.DataFrame()
        for aspect in self.aspects:
            df = df.combine_first(self.aspects[aspect].values_mean)
        return df["Pos - mean"]

    def values_std(self):
        df = pd.DataFrame()
        for aspect in self.aspects:
            df = df.combine_first(self.aspects[aspect].values_std)
        return df["Pos - mean"]

    @property
    def abnormality(self):
        if self._abnormality is None:
            df = pd.DataFrame()
            for aspect in self.aspects:
                df = df.combine_first(self.aspects[aspect].abnormality)
            self._abnormality = df
            return self._abnormality
        return self._abnormality

    def plot(self, data_type="diff", by_type_events=None, showlegend: bool = False, size_text: int =14,
             names_variables: dict = None):
        data_type_dict = {'std': "Standard deviation", "mean": "Mean"}
        data = []

        symbol = {"1-day maximum": 'circle', "Date of maximum": 'x', "High events count": 'triangle-up',
                  "High events duration": 'cross', "1-day minimum": 'circle', "Date of minimum": 'x',
                  "Low events count": 'triangle-up', "Low events duration": 'cross'}

        names = {"1-day maximum": "Magnitude", "Date of maximum": "Timing", "High events count": "Frequency",
                 "High events duration": "Duration", "1-day minimum": "Magnitude", "Date of minimum": "Timing",
                 "Low events count": "Frequency", "Low events duration": "Duration"}

        x = []
        error = []
        error_minus = []
        for i in self.aspects:
            for j in self.aspects[i].variables:
                graphs = GraphicsCha(obj_dhram=self.aspects[i].variables[j], data_type=data_type, xaxis="Variables",
                                     yaxis="Z-scores", size_text=size_text, names_variables=names_variables)
                try:
                    fig, df = graphs.plot(type="error_bar", by_type_events=by_type_events)
                except TypeError:
                    continue
                data_fig = fig["data"]
                error = error + list(df["Data"].values + df["Error"].values)
                error_minus = error_minus + list(df["Data"].values - df["Error_minus"].values)
                x = x + list(df["Variable"].values)
                data_fig[0]["name"] = names[j]
                data_fig[0]["marker"]["color"] = "black"
                data_fig[0]["marker"]["symbol"] = symbol[j]
                data_fig[0]["showlegend"] = False

                data_fig[1]["name"] = names[j]
                data_fig[1]["marker"]["color"] = "black"
                data_fig[1]["marker"]["symbol"] = symbol[j]
                data = data + [data_fig[0], data_fig[1]]
        layout = fig["layout"]
        layout['showlegend'] = showlegend
        layout["title"]["text"] = None
        fig = go.Figure(data=data, layout=layout)
        fig.add_trace(go.Scatter(
            x=x, y=[2] * len(x),
            mode='lines',
            line=dict(width=0.5, color='#2EFE2E'),
            stackgroup="one",
            showlegend=False,
        ))
        fig.add_trace(go.Scatter(
            x=x, y=[2] * len(x),
            mode='lines',
            line=dict(width=0.5, color='#F7FE2E'),
            stackgroup="one",
            showlegend=False
        ))
        fig.add_trace(go.Scatter(
            x=x, y=[2] * len(x),
            mode='lines',
            line=dict(width=0.5, color='#FAAC58'),
            stackgroup="one",
            showlegend=False
        ))
        if max(error) > 6:
            fig.add_trace(go.Scatter(
                x=x, y=[max(error) - 6] * len(x),
                mode='lines',
                line=dict(width=0.5, color='#FE2E2E'),
                stackgroup="one",
                showlegend=False
            ))
        fig.add_trace(go.Scatter(
            x=x, y=[-2] * len(x),
            mode='lines',
            line=dict(width=0.5, color='#2EFE2E'),
            stackgroup="two",
            showlegend=False
        ))
        fig.add_trace(go.Scatter(
            x=x, y=[-2] * len(x),
            mode='lines',
            line=dict(width=0.5, color='#F7FE2E'),
            stackgroup="two",
            showlegend=False
        ))
        fig.add_trace(go.Scatter(
            x=x, y=[-2] * len(x),
            mode='lines',
            line=dict(width=0.5, color='#FAAC58'),
            stackgroup="two",
            showlegend=False
        ))
        error_minus_not_null = [x for x in error_minus if str(x) != 'nan']
        if min(error_minus_not_null) < -6:
            fig.add_trace(go.Scatter(
                x=x, y=[min(error_minus_not_null)+6] * len(x),
                mode='lines',
                line=dict(width=0.5, color='#FE2E2E'),
                stackgroup="two",
                showlegend=False
            ))
        return fig, data


class EraAspect:

    def __init__(self, name):
        self.name = name
        self._abnormality = None
        self._variables: dict = {}
        self._list_name_variables = []

    @property
    def variables(self):
        return self._variables

    @variables.setter
    def variables(self, variable):
        self._list_name_variables.append(variable.name)
        self._variables[variable.name] = variable

    @property
    def values_mean(self):
        df = pd.DataFrame()
        for i in self.variables:
            df = df.combine_first(self.variables[i].value_mean)
        return df.reindex(self._list_name_variables)

    @property
    def values_std(self):
        df = pd.DataFrame()
        for i in self.variables:
            df = df.combine_first(self.variables[i].value_std)
        return df.reindex(self._list_name_variables)

    @property
    def abnormality(self):
        if self._abnormality is None:
            df = pd.DataFrame()
            for i in self.variables:
                df = df.combine_first(self.variables[i].abnormality)
            self._abnormality = df.reindex(self._list_name_variables)

        return self._abnormality

    def plot(self, data_type="mean"):
        graphs = GraphicsCha(obj_dhram=self, data_type=data_type, xaxis="Variable", yaxis="Abnormality")
        fig, data = graphs.plot(type="error_bar")
        return fig, data


class EraVariable:

    def __init__(self, variable_pre, variable_pos, interval: int, m: int):
        self.name = variable_pre.name
        self.data_pre = variable_pre.data
        self.data_pos = variable_pos.data
        self.data_pos_nulls = self.data_pos
        self.interval = [((100 - interval) / 2) / 100, 1 - ((100 - interval) / 2) / 100]
        self.sample_pre = Bootstrap(data=self.data_pre, m=m, name=self.name)
        self.sample_pos = Bootstrap(data=self.data_pos, m=m, name=self.name)
        self.sample_diff = Bootstrap(data=self.data_pre-self.data_pos, m=m, name=self.name)

    def __calc_value(self, dist_pre, dist_pos, dist_diff):
        value = pd.DataFrame(columns=["Pre - 2_5", "Pre - 97_5", "Pos - 2_5", "Pos - 97_5", "Pre - mean", "Pos - mean"],
                             dtype='float64')
        confidence_intervals_pre = dist_pre.data.quantile(self.interval)
        confidence_intervals_pos = dist_pos.data.quantile(self.interval)
        confidence_intervals_diff = dist_diff.data.quantile(self.interval)
        zscore_pre_2_5 = dist_pre.z_score(confidence_intervals_pre[self.interval[0]])
        zscore_pos_2_5 = dist_pre.z_score(confidence_intervals_pos[self.interval[0]])
        zscore_pre_97_5 = dist_pre.z_score(confidence_intervals_pre[self.interval[1]])
        zscore_pos_97_5 = dist_pre.z_score(confidence_intervals_pos[self.interval[1]])
        zscore_pre_mean = dist_pre.z_score(dist_pre.data.mean())
        zscore_pos_mean = dist_pre.z_score(dist_pos.data.mean())

        value.at[self.name, "Pre - 2_5"] = zscore_pre_2_5
        value.at[self.name, "Pre - mean"] = zscore_pre_mean
        value.at[self.name, "Pre - 97_5"] = zscore_pre_97_5

        value.at[self.name, "Pos - 2_5"] = zscore_pos_2_5
        value.at[self.name, "Pos - mean"] = zscore_pos_mean
        value.at[self.name, "Pos - 97_5"] = zscore_pos_97_5
        return value

    @property
    def value_mean(self):

        dist_pre = Normal(data=self.sample_pre.mean())
        dist_pos = Normal(data=self.sample_pos.mean())
        dist_diff = Normal(data=self.sample_diff.mean())

        value = self.__calc_value(dist_pre=dist_pre, dist_pos=dist_pos, dist_diff=dist_diff)
        return value

    @property
    def value_std(self):

        dist_pre = Normal(data=self.sample_pre.std())
        dist_pos = Normal(data=self.sample_pos.std())
        dist_diff = Normal(data=self.sample_diff.std())

        value = self.__calc_value(dist_pre=dist_pre, dist_pos=dist_pos, dist_diff=dist_diff)

        return value

    @staticmethod
    def __calc_abnormality(pos_2_5, pos_97_5):
        if abs(pos_2_5) < abs(pos_97_5):
            value_pos = pos_2_5
        else:
            value_pos = pos_97_5

        multi_diff = (abs(abs(value_pos) - 2) / 2) * value_pos/abs(value_pos)
        return multi_diff

    @property
    def abnormality(self) -> pd.DataFrame:

        diff_df = pd.DataFrame(columns=["Abnormality_mean", "Abnormality_std"])

        z_score_mean = self.value_mean
        z_score_std = self.value_std
        pos_mean_2_5, pos_std_2_5 = z_score_mean["Pos - 2_5"].values[0], z_score_std["Pos - 2_5"].values[0]
        pos_mean_97_5, pos_std_97_5 = z_score_mean["Pos - 97_5"].values[0], z_score_std["Pos - 97_5"].values[0]

        value_mean_pos = self.__calc_abnormality(pos_mean_2_5, pos_mean_97_5)
        value_std_pos = self.__calc_abnormality(pos_std_2_5, pos_std_97_5)

        diff_df.at[self.name, "Abnormality_mean"] = value_mean_pos
        diff_df.at[self.name, "Abnormality_std"] = value_std_pos

        return diff_df

    def plot(self, color={"pre": "blue", "pos": "red"}, type="mean"):
        """
        @type color: dict
        """
        if type == "mean":
            fig_obs, data_obs = GraphicsCha(obj_dhram=self, color=color, name=self.name).plot(type="point")
            fig_nat, data_nat = GraphicsCha(obj_dhram=self, color=color, name=self.name).plot(type="point")
        elif type == "std":
            fig_obs, data_obs = GraphicsCha(obj_dhram=self, color=color, name=self.name).plot(type="point")
            fig_nat, data_nat = GraphicsCha(obj_dhram=self, color=color, name=self.name).plot(type="point")
        else:
            raise AttributeError("Type Error")

        data = data_obs + data_nat
        fig = dict(data=data, layout=fig_nat['layout'])

        return fig, data

    def __str__(self):
        return self.name



import pandas as pd


class Bootstrap:

    def __init__(self, data: pd.Series, m: int, name):
        self._sample = None
        self.data = data
        self.n = len(data)
        self.m = m
        self.name = name

    @property
    def sample(self) -> pd.DataFrame:
        if self._sample is None:
            sample = []
            for i in range(self.m):
                sample.append(pd.Series(data=self.data.sample(n=self.n, replace=True).values, name=i+1))
            self._sample = pd.DataFrame(sample).T
        return self._sample

    def mean(self):
        series_mean = self.sample.mean()
        series_mean.name = self.name
        return series_mean

    def std(self):
        return self.sample.std()

    def __str__(self):
        return self.sample.__repr__()

    def __getitem__(self, item):
        return self.sample[item]



from abc import ABCMeta, abstractmethod

import plotly.graph_objs as go
import plotly.express as px
import pandas as pd


class Graphics(metaclass=ABCMeta):

    def __init__(self, obj, color=None, width=None, height=None, size_text=14, xaxis=None,
                 yaxis=None, name=None):
        self.obj = obj
        self.name = name
        self.width = width
        self.height = height
        self.size_text = size_text
        self.color = color
        self.xaxis = xaxis
        self.yaxis = yaxis

    @abstractmethod
    def plot(self):
        pass

    def layout(self):
        bandxaxis = go.layout.XAxis(title=self.xaxis)
        bandyaxis = go.layout.YAxis(title=self.yaxis)
        layout = dict(
            title=dict(text=self.name, x=0.5, xanchor='center', y=0.95, yanchor='top',
                       font=dict(family='Courier New, monospace', size=self.size_text + 10)),
            xaxis=bandxaxis,
            yaxis=bandyaxis,
            width=self.width, height=self.height,
            font=dict(family='Courier New, monospace', size=self.size_text, color='rgb(0,0,0)'),
            showlegend=True, plot_bgcolor='#FFFFFF', paper_bgcolor='#FFFFFF')

        return layout


class GraphicsRVA(Graphics):

    def __init__(self, data_variable, line=None, color=None, width=None, height=None, size_text=14, xaxis=None,
                 yaxis=None):
        super().__init__(obj=data_variable, color=color, width=width, height=height,
                         size_text=size_text, xaxis=xaxis, yaxis=yaxis)
        self.line = line

    def _plot_rva(self, line):
        dash = {"upper_line": "dot", "lower_line": "dash"}
        line_graph = go.Scatter(x=self.data.index,
                                y=[self.line[line][self.name]] * len(self.data.index),
                                name=line.title(),
                                mode='lines',
                                line=dict(width=1, dash=dash[line], color='rgb(0, 0, 0)'),
                                opacity=1, connectgaps=False)
        return line_graph

    def _plot_iha(self):
        point = go.Scatter(x=self.data.index,
                           y=self.data.loc[self.data.index].values,
                           name=self.status.title()+'-impacto',
                           mode='markers+lines',
                           marker=dict(size=8, color=self.color[self.status], line=dict(width=1,
                                                                                        color=self.color[self.status])))

        return point

    def plot(self):
        layout = self.layout()

        data = list()
        data.append(self._plot_iha())
        try:
            for i in self.line:
                if i != 'median_line':
                    data.append(self._plot_rva(i))
        except:
            pass

        fig = dict(data=data, layout=layout)
        return fig, data


class GraphicsCha(Graphics):

    def __init__(self, obj_dhram, color=None, width=None, height=None, size_text=14, xaxis=None, yaxis=None, name=None,
                 data_type=None, names_variables: dict = None):
        super().__init__(obj=obj_dhram, color=color, width=width, height=height, size_text=size_text, xaxis=xaxis,
                         yaxis=yaxis, name=name)
        self.data_type = data_type
        self.names_variables = names_variables

    def plot(self, type="error_bar", by_type_events=None):
        layout = self.layout()

        data = list()
        if type == "point":
            data.append(self._point_simulation())
            fig = dict(data=data, layout=layout)
            return fig, data
        elif type == "box":
            pass
        elif type == "violin":
            pass
        elif type == "error_bar":
            return self._error_bar(by_type_events)
        else:
            raise TypeError

    def _box(self):
        pass

    def _error_bar(self, by_type_events):
        dic = {"Data": [], "Status": [], "Variable": [], "Error": [], "Error_minus": [], "Text": []}
        events_high = ["1-day maximum", "Date of maximum", "High events count", "High events duration"]
        events_low = ["1-day minimum", "Date of minimum", "Low events count", "Low events duration"]

        # names = {"1-day maximum": "1-day", "Date of maximum": "Date", "High events count": "Count",
        #          "High events duration": "Duration", "1-day minimum": "1-day", "Date of minimum": "Date",
        #          "Low events count": "Count", "Low events duration": "Duration"}

        dict_events = {"High": events_high, "Low": events_low}
        if self.data_type == "mean":
            if self.obj.name in dict_events[by_type_events]:
                variable_diff = self.obj.value_mean
                variable = {"Diff": variable_diff}
            else:
                return None
        elif self.data_type == "std":
            if self.obj.name in dict_events[by_type_events]:
                variable_diff = self.obj.value_std
                variable = {"Diff": variable_diff}
            else:
                return None
        else:
            if self.obj.name in dict_events[by_type_events]:
                variable_diff_mean = self.obj.value_mean
                variable_diff_std = self.obj.value_std
                variable = {"Mean": variable_diff_mean, "Std": variable_diff_std}
            else:
                return None

        for j in variable:
            dic["Data"] = dic["Data"] + [variable[j]["Pos - mean"].values[0]]
            dic["Status"] = dic["Status"] + [j]
            dic["Variable"] = dic["Variable"] + [self.names_variables[self.obj.name] + "("+j+")"
                                                 if self.obj.name in self.names_variables.keys() else self.obj.name]
            dic["Error"] = dic["Error"] + [variable[j]["Pos - 97_5"].values[0] - variable[j]["Pos - mean"].values[0]]
            dic["Error_minus"] = dic["Error_minus"] + [variable[j]["Pos - mean"].values[0] - variable[j]["Pos - 2_5"].values[0]]

            dic["Text"] = dic["Text"] + [f"{(self.obj.interval[1] - self.obj.interval[0]) * 100}% "
                                         f"({round(variable[j]['Pos - 2_5'].values[0], 2)}"
                                         f"; {round(variable[j]['Pos - 97_5'].values[0], 2)})"]
        df = pd.DataFrame(dic)
        fig = px.scatter(df, x="Variable", y="Data", color="Status", error_y="Error", error_y_minus="Error_minus",
                         text="Text", labels={"Text": "Variable", "Data": ""},
                         hover_data={"Data": ":.2f"})

        fig.update_traces(mode="markers")
        fig.layout = self.layout()
        fig.layout.title.text = f"{self.obj.name} - {self.data_type.title()}"
        return fig, df



import calendar as cal
import pandas as pd

from hidrocomp.series.exceptions import StationError
from hidrocomp.statistic.pearson3 import Pearson3
from hidrocomp.series.series_build import SeriesBuild
from hidrocomp.series.partial import Partial
from hidrocomp.series.maximum import Maximum
from hidrocomp.series.minimum import Minimum
from hidrocomp.eflow import IHA
from hidrocomp.graphics import RatingCurve, HydrogramYear, HydrogramClean


class Flow(SeriesBuild):

    type_data = 'FLUVIOMÉTRICO'

    def __init__(self, data=None, path_file=None, station=None, source=None, *args, **kwargs):
        super().__init__(data=data, path=path_file, station=station, source=source, type_data=self.type_data, *args,
                         **kwargs)
        self.__month_num_flood = None
        self.__month_abr_flood = None
        self.__month_num_drought = None
        self.__month_abr_drought = None

    def _month_start_year_hydrologic(self):
        if self.__month_num_flood is None:
            if self.station is None:
                raise TypeError("Define a station!")
            else:
                data = pd.DataFrame(self.data[self.station])
                mean_month = pd.DataFrame([data.loc[data.index.month == i].mean() for i in range(1, 13)])
                month_start_year_hydrologic = 1 + mean_month.idxmin().values[0]
                month_start_year_hydrologic_abr_flood = cal.month_abbr[month_start_year_hydrologic].upper()

                self.__month_num_flood = month_start_year_hydrologic
                self.__month_abr_flood = 'AS-%s' % month_start_year_hydrologic_abr_flood

                self.__month_num_drought = month_start_year_hydrologic - 6
                self.__month_abr_drought = 'AS-%s' % cal.month_abbr[self.__month_num_drought].upper()
        else:
            if self.__month_num_flood > 6:
                self.__month_num_drought = self.__month_num_flood - 6
            else:
                self.__month_num_drought = self.__month_num_flood + 6

            self.__month_abr_flood = 'AS-%s' % cal.month_abbr[self.__month_num_flood].upper()
            self.__month_abr_drought = 'AS-%s' % cal.month_abbr[self.__month_num_drought].upper()

        return self.__month_num_flood, self.__month_abr_flood, self.__month_num_drought, self.__month_abr_drought

    @property
    def month_num_flood(self) -> int:
        return self._month_start_year_hydrologic()[0]

    @month_num_flood.setter
    def month_num_flood(self, month: int):
        self.__month_num_flood = month

    @property
    def month_abr_flood(self) -> str:
        return self._month_start_year_hydrologic()[1]

    @property
    def month_num_drought(self) -> int:
        return self._month_start_year_hydrologic()[2]

    @property
    def month_abr_drought(self) -> str:
        return self._month_start_year_hydrologic()[3]

    def minimum(self):
        minimum = Minimum(obj=self, station=self.station)

        return minimum

    def maximum(self):
        maximum = Maximum(obj=self, station=self.station)

        return maximum

    def partial(self, type_threshold, type_event, type_criterion, value_threshold, **kwargs):
        partial = Partial(station=self.station, obj=self, type_threshold=type_threshold, type_event=type_event,
                          type_criterion=type_criterion, value_threshold=value_threshold, **kwargs)

        return partial

    def simulation_withdraw(self, criterion, rate, months=None, value=None):
        if type(months) is not list and months is not None:
            raise TypeError

        if criterion == 'q90':
            data = self.data.copy()
            if value is None:
                withdraw = (self.quantile(percentile=0.1) * (rate / 100)).values[0]
            else:
                withdraw = (value * (rate / 100)).values[0]

            if months is None:
                name = "withdraw_{}_{}_{}".format(rate, criterion, 'all')
                data.rename(columns={self.station: name}, inplace=True)
                data = data - withdraw
                data.loc[data[name] < 0, name] = 0
                return data
            else:
                name = "withdraw_{}_{}_{}".format(rate, criterion, 'months')
                data.rename(columns={self.station: name}, inplace=True)
                for i in self.data[self.station].index:
                    if i.month in months:
                        data.at[i, name] = self.data[self.station][i] - withdraw
                        data.loc[data[name] < 0, name] = 0
                return data
        else:
            return None

    def hydrogram(self, title, threshold=None, save=False, width=None, height=None, y_title='Flow (m³/s)',
                  x_title='Date', size_text=16, color=None):
        if self.station is None:
            hydrogram = HydrogramClean(self.data, threshold=threshold, width=width, height=height, size_text=size_text,
                                       title=title, y_title=y_title, x_title=x_title, color=color)
            fig, data = hydrogram.plot()
        else:
            hydrogram = HydrogramClean(self.data[self.station], threshold=threshold, width=width, height=height,
                                       size_text=size_text, title=title, y_title=y_title, x_title=x_title, color=color)
            fig, data = hydrogram.plot()
        return fig, data

    def iha(self, status=None, date_start: str = None, date_end: str = None, statistic="no-parametric",
            central_metric="mean", month_water: int = None, variation_metric: str = "std", type_threshold="stationary",
            type_criterion: str = None, threshold_high: float = None, threshold_low: float = None,
            aspects: list = None, magnitude: list = None, magnitude_and_duration: list = None, timing: list = None,
            frequency_and_duration: list = None, rate_and_frequency: list = None, **kwargs) -> IHA:
        """
        @param rate_and_frequency:
        @param frequency_and_duration:
        @param timing:
        @param magnitude_and_duration:
        @param magnitude:
        @param aspects:
        @param status:
        @param date_start:
        @param date_end:
        @param statistic:
        @param central_metric:
        @param month_water:
        @param variation_metric:
        @param type_threshold:
        @param type_criterion:
        @param threshold_high:
        @param threshold_low:
        @return:
        """
        iha = IHA(flow=self, month_water=month_water, status=status, date_start=date_start, date_end=date_end,
                  statistic=statistic, central_metric=central_metric, variation_metric=variation_metric,
                  type_threshold=type_threshold, type_criterion=type_criterion, threshold_high=threshold_high,
                  threshold_low=threshold_low, aspects=aspects, magnitude=magnitude, timing=timing,
                  frequency_and_duration=frequency_and_duration, rate_and_frequency=rate_and_frequency,
                  magnitude_and_duration=magnitude_and_duration, **kwargs)
        return iha

    def power_energy(self, efficiency: int, gravity: float, hydraulic_head: float, station: str = None) -> pd.Series:
        if len(self.columns) > 1:
            if station is None:
                raise AttributeError("Station is None")
            else:
                data = self.data[station]
        else:
            data = self.data
        const = efficiency * gravity * hydraulic_head
        pot = data.multiply(const)
        return pot

    def hydrogram_year(self, title="", threshold=None, width=None, height=None, size_text=16):
        idx = [i for i in self.data.index if i.month == 2 and i.day == 29]
        data = self.data.drop(index=idx)
        data = data.groupby(pd.Grouper(freq=self.month_abr_flood))
        hydrogram = HydrogramYear(data=data, threshold=threshold, width=width, height=height, title=title,
                                  size_text=size_text)
        fig, data = hydrogram.plot()
        return fig, data

    def rating_curve(self, title=None, width=None, height=None, size_text=16):
        if self.station is None:
            raise StationError
        permanence = RatingCurve(self.data[self.station], width=width, height=height, size_text=size_text,
                                 title=title)
        fig, data = permanence.plot()
        return fig, data

    def get_month_name(self):
        months = {1: 'Janeiro', 2: 'Fevereiro', 3: 'Março', 4: 'Abril', 5: 'Maio', 6: 'Junho', 7: 'Julho', 8: 'Agosto',
                  9: 'Setembro', 10: 'Outubro', 11: 'Novembro', 12: 'Dezembro'}
        return months[self.month_num_flood]

    def flow_min(self, method):
        if method == 'q90':
            return self.quantile(0.1)
        elif method == 'q95':
            return self.quantile(0.05)
        elif method == 'q710':
            qmin = self.data.rolling(7).mean().groupby(pd.Grouper(freq='AS-JAN')).min()
            prop = 1 / 10
            dist = Pearson3(qmin[self.station].values)
            dist.mml()
            return dist.values(prop)
